@@ -1,96 +1,51 @@
 import asyncio
 import contextlib
 import errno
+import platform
 import select
 import sys
 import traceback
 import warnings
 import weakref
-import platform
+from collections.abc import Mapping
 
 import psycopg2
-from psycopg2.extensions import (
-    POLL_OK, POLL_READ, POLL_WRITE, POLL_ERROR)
-from psycopg2 import extras
+import psycopg2.extensions
+import psycopg2.extras
 
 from .cursor import Cursor
-from .utils import _ContextManager, PY_35, create_future
-
+from .utils import _ContextManager, create_completed_future, get_running_loop
 
 __all__ = ('connect',)
 
-
 TIMEOUT = 60.0
-PY_341 = sys.version_info >= (3, 4, 1)
 
 # Windows specific error code, not in errno for some reason, and doesnt map
 # to OSError.errno EBADF
 WSAENOTSOCK = 10038
 
 
-@asyncio.coroutine
-def _enable_hstore(conn):
-    cur = yield from conn.cursor()
-    yield from cur.execute("""\
-        SELECT t.oid, typarray
-        FROM pg_type t JOIN pg_namespace ns
-            ON typnamespace = ns.oid
-        WHERE typname = 'hstore';
-        """)
-    rv0, rv1 = [], []
-    for oids in (yield from cur.fetchall()):
-        if isinstance(oids, dict):
-            rv0.append(oids['oid'])
-            rv1.append(oids['typarray'])
-        else:
-            rv0.append(oids[0])
-            rv1.append(oids[1])
-
-    cur.close()
-    return tuple(rv0), tuple(rv1)
-
-
-def connect(dsn=None, *, timeout=TIMEOUT, loop=None, enable_json=True,
+def connect(dsn=None, *, timeout=TIMEOUT, enable_json=True,
             enable_hstore=True, enable_uuid=True, echo=False, **kwargs):
     """A factory for connecting to PostgreSQL.
 
     The coroutine accepts all parameters that psycopg2.connect() does
-    plus optional keyword-only `loop` and `timeout` parameters.
+    plus optional keyword-only `timeout` parameters.
 
     Returns instantiated Connection object.
 
     """
-    coro = _connect(dsn=dsn, timeout=timeout, loop=loop,
-                    enable_json=enable_json, enable_hstore=enable_hstore,
-                    enable_uuid=enable_uuid, echo=echo, **kwargs)
+    coro = Connection(
+        dsn, timeout, bool(echo),
+        enable_hstore=enable_hstore,
+        enable_uuid=enable_uuid,
+        enable_json=enable_json,
+        **kwargs
+    )
+
     return _ContextManager(coro)
 
 
-@asyncio.coroutine
-def _connect(dsn=None, *, timeout=TIMEOUT, loop=None, enable_json=True,
-             enable_hstore=True, enable_uuid=True, echo=False, **kwargs):
-    if loop is None:
-        loop = asyncio.get_event_loop()
-
-    waiter = create_future(loop)
-    conn = Connection(dsn, loop, timeout, waiter, bool(echo), **kwargs)
-    try:
-        yield from conn._poll(waiter, timeout)
-    except Exception:
-        conn.close()
-        raise
-    if enable_json:
-        extras.register_default_json(conn._conn)
-    if enable_uuid:
-        extras.register_uuid(conn_or_curs=conn._conn)
-    if enable_hstore:
-        oids = yield from _enable_hstore(conn)
-        if oids is not None:
-            oid, array_oid = oids
-            extras.register_hstore(conn._conn, oid=oid, array_oid=array_oid)
-    return conn
-
-
 def _is_bad_descriptor_error(os_error):
     if platform.system() == 'Windows':  # pragma: no cover
         return os_error.winerror == WSAENOTSOCK
@@ -108,25 +63,32 @@ class Connection:
 
     _source_traceback = None
 
-    def __init__(self, dsn, loop, timeout, waiter, echo, **kwargs):
-        self._loop = loop
-        self._conn = psycopg2.connect(dsn, async_=True, **kwargs)
+    def __init__(
+            self, dsn, timeout, echo,
+            *, enable_json=True, enable_hstore=True,
+            enable_uuid=True, **kwargs
+    ):
+        self._enable_json = enable_json
+        self._enable_hstore = enable_hstore
+        self._enable_uuid = enable_uuid
+        self._loop = get_running_loop()
+        self._waiter = self._loop.create_future()
+
+        kwargs['async_'] = kwargs.pop('async', True)
+        self._conn = psycopg2.connect(dsn, **kwargs)
+
         self._dsn = self._conn.dsn
         assert self._conn.isexecuting(), "Is conn an async at all???"
         self._fileno = self._conn.fileno()
         self._timeout = timeout
         self._last_usage = self._loop.time()
-        self._waiter = waiter
         self._writing = False
-        self._cancelling = False
-        self._cancellation_waiter = None
         self._echo = echo
-        self._conn_cursor = None
-        self._notifies = asyncio.Queue(loop=loop)
+        self._notifies = asyncio.Queue()
         self._weakref = weakref.ref(self)
         self._loop.add_reader(self._fileno, self._ready, self._weakref)
 
-        if loop.get_debug():
+        if self._loop.get_debug():
             self._source_traceback = traceback.extract_stack(sys._getframe(1))
 
     @staticmethod
@@ -173,94 +135,70 @@ class Connection:
                 if waiter is not None and not waiter.done():
                     waiter.set_exception(
                         psycopg2.OperationalError("Connection closed"))
-            if state == POLL_OK:
+            if state == psycopg2.extensions.POLL_OK:
                 if self._writing:
                     self._loop.remove_writer(self._fileno)
                     self._writing = False
                 if waiter is not None and not waiter.done():
                     waiter.set_result(None)
-            elif state == POLL_READ:
+            elif state == psycopg2.extensions.POLL_READ:
                 if self._writing:
                     self._loop.remove_writer(self._fileno)
                     self._writing = False
-            elif state == POLL_WRITE:
+            elif state == psycopg2.extensions.POLL_WRITE:
                 if not self._writing:
                     self._loop.add_writer(self._fileno, self._ready, weak_self)
                     self._writing = True
-            elif state == POLL_ERROR:
+            elif state == psycopg2.extensions.POLL_ERROR:
                 self._fatal_error("Fatal error on aiopg connection: "
                                   "POLL_ERROR from underlying .poll() call")
             else:
-                self._fatal_error("Fatal error on aiopg connection: "
-                                  "unknown answer {} from underlying "
-                                  ".poll() call"
-                                  .format(state))
+                self._fatal_error(f"Fatal error on aiopg connection: "
+                                  f"unknown answer {state} from underlying "
+                                  f".poll() call")
 
     def _fatal_error(self, message):
         # Should be called from exception handler only.
         self._loop.call_exception_handler({
             'message': message,
             'connection': self,
-            })
+        })
         self.close()
         if self._waiter and not self._waiter.done():
             self._waiter.set_exception(psycopg2.OperationalError(message))
 
     def _create_waiter(self, func_name):
         if self._waiter is not None:
-            if self._cancelling:
-                if not self._waiter.done():
-                    raise RuntimeError('%s() called while connection is '
-                                       'being cancelled' % func_name)
-            else:
-                raise RuntimeError('%s() called while another coroutine is '
-                                   'already waiting for incoming '
-                                   'data' % func_name)
-        self._waiter = create_future(self._loop)
+            raise RuntimeError(f'{func_name}() called while another coroutine '
+                               f'is already waiting for incoming data')
+        self._waiter = self._loop.create_future()
         return self._waiter
 
-    @asyncio.coroutine
-    def _poll(self, waiter, timeout):
+    async def _poll(self, waiter, timeout):
         assert waiter is self._waiter, (waiter, self._waiter)
         self._ready(self._weakref)
 
-        @asyncio.coroutine
-        def cancel():
-            self._waiter = create_future(self._loop)
-            self._cancelling = True
-            self._cancellation_waiter = self._waiter
-            self._conn.cancel()
-            if not self._conn.isexecuting():
-                return
-            try:
-                yield from asyncio.wait_for(self._waiter, timeout,
-                                            loop=self._loop)
-            except psycopg2.extensions.QueryCanceledError:
-                pass
-            except asyncio.TimeoutError:
-                self._close()
-
         try:
-            yield from asyncio.wait_for(self._waiter, timeout, loop=self._loop)
+            await asyncio.wait_for(self._waiter, timeout)
         except (asyncio.CancelledError, asyncio.TimeoutError) as exc:
-            yield from asyncio.shield(cancel(), loop=self._loop)
+            await asyncio.shield(self.close())
             raise exc
-        except psycopg2.extensions.QueryCanceledError:
+        except psycopg2.extensions.QueryCanceledError as exc:
+            self._loop.call_exception_handler({
+                'message': exc.pgerror,
+                'exception': exc,
+                'future': self._waiter,
+            })
             raise asyncio.CancelledError
         finally:
-            if self._cancelling:
-                self._cancelling = False
-                if self._waiter is self._cancellation_waiter:
-                    self._waiter = None
-                self._cancellation_waiter = None
-            else:
-                self._waiter = None
+            self._waiter = None
 
     def _isexecuting(self):
         return self._conn.isexecuting()
 
     def cursor(self, name=None, cursor_factory=None,
-               scrollable=None, withhold=False, timeout=None):
+               scrollable=None, withhold=False, timeout=None,
+               isolation_level=None):
         """A coroutine that returns a new cursor object using the connection.
 
         *cursor_factory* argument can be used to create non-standard
@@ -270,45 +208,32 @@ class Connection:
         *name*, *scrollable* and *withhold* parameters are not supported by
         psycopg in asynchronous mode.
 
-        NOTE: as of [TODO] any previously created created cursor from this
-            connection will be closed
         """
-        self.close_cursor()
 
         self._last_usage = self._loop.time()
         coro = self._cursor(name=name, cursor_factory=cursor_factory,
                             scrollable=scrollable, withhold=withhold,
-                            timeout=timeout)
+                            timeout=timeout, isolation_level=isolation_level)
         return _ContextManager(coro)
 
-    def cursor_created(self, cursor):
-        if self._conn_cursor and not self._conn_cursor.closed:
-            raise Exception("You can only have one cursor per connection")
+    async def _cursor(self, name=None, cursor_factory=None,
+                      scrollable=None, withhold=False, timeout=None,
+                      isolation_level=None):
 
-        self._conn_cursor = cursor
-
-    def cursor_closed(self, cursor):
-        if cursor != self._conn_cursor:
-            raise Exception("You can only have one cursor per connection")
-
-        self._conn_cursor = None
-
-    @asyncio.coroutine
-    def _cursor(self, name=None, cursor_factory=None,
-                scrollable=None, withhold=False, timeout=None):
         if timeout is None:
             timeout = self._timeout
 
-        impl = yield from self._cursor_impl(name=name,
-                                            cursor_factory=cursor_factory,
-                                            scrollable=scrollable,
-                                            withhold=withhold)
-        cursor = Cursor(self, impl, timeout, self._echo)
+        impl = await self._cursor_impl(name=name,
+                                       cursor_factory=cursor_factory,
+                                       scrollable=scrollable,
+                                       withhold=withhold)
+        cursor = Cursor(
+            self, impl, timeout, self._echo, isolation_level
+        )
         return cursor
 
-    @asyncio.coroutine
-    def _cursor_impl(self, name=None, cursor_factory=None,
-                     scrollable=None, withhold=False):
+    async def _cursor_impl(self, name=None, cursor_factory=None,
+                           scrollable=None, withhold=False):
         if cursor_factory is None:
             impl = self._conn.cursor(name=name,
                                      scrollable=scrollable, withhold=withhold)
@@ -327,7 +252,6 @@ class Connection:
                 self._writing = False
                 self._loop.remove_writer(self._fileno)
 
-        self.close_cursor()
         self._conn.close()
 
         if self._waiter is not None and not self._waiter.done():
@@ -336,13 +260,7 @@ class Connection:
 
     def close(self):
         self._close()
-        ret = create_future(self._loop)
-        ret.set_result(None)
-        return ret
-
-    def close_cursor(self):
-        if self._conn_cursor:
-            self._conn_cursor.close()
+        return create_completed_future(self._loop)
 
     @property
     def closed(self):
@@ -359,65 +277,44 @@ class Connection:
         """Underlying psycopg connection object, readonly"""
         return self._conn
 
-    @asyncio.coroutine
-    def commit(self):
+    async def commit(self):
         raise psycopg2.ProgrammingError(
             "commit cannot be used in asynchronous mode")
 
-    @asyncio.coroutine
-    def rollback(self):
+    async def rollback(self):
         raise psycopg2.ProgrammingError(
             "rollback cannot be used in asynchronous mode")
 
     # TPC
 
-    @asyncio.coroutine
-    def xid(self, format_id, gtrid, bqual):
+    async def xid(self, format_id, gtrid, bqual):
         return self._conn.xid(format_id, gtrid, bqual)
 
-    @asyncio.coroutine
-    def tpc_begin(self, xid=None):
+    async def tpc_begin(self, xid=None):
         raise psycopg2.ProgrammingError(
             "tpc_begin cannot be used in asynchronous mode")
 
-    @asyncio.coroutine
-    def tpc_prepare(self):
+    async def tpc_prepare(self):
         raise psycopg2.ProgrammingError(
             "tpc_prepare cannot be used in asynchronous mode")
 
-    @asyncio.coroutine
-    def tpc_commit(self, xid=None):
+    async def tpc_commit(self, xid=None):
         raise psycopg2.ProgrammingError(
             "tpc_commit cannot be used in asynchronous mode")
 
-    @asyncio.coroutine
-    def tpc_rollback(self, xid=None):
+    async def tpc_rollback(self, xid=None):
         raise psycopg2.ProgrammingError(
             "tpc_rollback cannot be used in asynchronous mode")
 
-    @asyncio.coroutine
-    def tpc_recover(self):
+    async def tpc_recover(self):
         raise psycopg2.ProgrammingError(
             "tpc_recover cannot be used in asynchronous mode")
 
-    @asyncio.coroutine
-    def cancel(self):
-        """Cancel the current database operation."""
-        if self._waiter is None:
-            return
+    async def cancel(self):
+        raise psycopg2.ProgrammingError(
+            "cancel cannot be used in asynchronous mode")
 
-        @asyncio.coroutine
-        def cancel():
-            self._conn.cancel()
-            try:
-                yield from self._waiter
-            except psycopg2.extensions.QueryCanceledError:
-                pass
-
-        yield from asyncio.shield(cancel(), loop=self._loop)
-
-    @asyncio.coroutine
-    def reset(self):
+    async def reset(self):
         raise psycopg2.ProgrammingError(
             "reset cannot be used in asynchronous mode")
 
@@ -431,9 +328,8 @@ class Connection:
         """
         return self._dsn
 
-    @asyncio.coroutine
-    def set_session(self, *, isolation_level=None, readonly=None,
-                    deferrable=None, autocommit=None):
+    async def set_session(self, *, isolation_level=None, readonly=None,
+                          deferrable=None, autocommit=None):
         raise psycopg2.ProgrammingError(
             "set_session cannot be used in asynchronous mode")
 
@@ -456,8 +352,7 @@ class Connection:
         """
         return self._conn.isolation_level
 
-    @asyncio.coroutine
-    def set_isolation_level(self, val):
+    async def set_isolation_level(self, val):
         """Transaction isolation level.
 
         The only allowed value is ISOLATION_LEVEL_READ_COMMITTED.
@@ -470,8 +365,7 @@ class Connection:
         """Client encoding for SQL operations."""
         return self._conn.encoding
 
-    @asyncio.coroutine
-    def set_client_encoding(self, val):
+    async def set_client_encoding(self, val):
         self._conn.set_client_encoding(val)
 
     @property
@@ -484,18 +378,15 @@ class Connection:
         """The default cursor factory used by .cursor()."""
         return self._conn.cursor_factory
 
-    @asyncio.coroutine
-    def get_backend_pid(self):
+    async def get_backend_pid(self):
         """Returns the PID of the backend server process."""
         return self._conn.get_backend_pid()
 
-    @asyncio.coroutine
-    def get_parameter_status(self, parameter):
+    async def get_parameter_status(self, parameter):
         """Look up a current parameter setting of the server."""
         return self._conn.get_parameter_status(parameter)
 
-    @asyncio.coroutine
-    def get_transaction_status(self):
+    async def get_transaction_status(self):
         """Return the current session transaction status as an integer."""
         return self._conn.get_transaction_status()
 
@@ -514,8 +405,7 @@ class Connection:
         """A read-only integer representing the status of the connection."""
         return self._conn.status
 
-    @asyncio.coroutine
-    def lobject(self, *args, **kwargs):
+    async def lobject(self, *args, **kwargs):
         raise psycopg2.ProgrammingError(
             "lobject cannot be used in asynchronous mode")
 
@@ -534,33 +424,86 @@ class Connection:
         """Return echo mode status."""
         return self._echo
 
-    if PY_341:  # pragma: no branch
-        def __del__(self):
-            try:
-                _conn = self._conn
-            except AttributeError:
-                return
-            if _conn is not None and not _conn.closed:
-                self.close()
-                warnings.warn("Unclosed connection {!r}".format(self),
-                              ResourceWarning)
+    def __repr__(self):
+        return (
+            f'<'
+            f'{type(self).__module__}::{type(self).__name__} '
+            f'isexecuting={self._isexecuting()}, '
+            f'closed={self.closed}, '
+            f'echo={self.echo}, '
+            f'>'
+        )
 
-                context = {'connection': self,
-                           'message': 'Unclosed connection'}
-                if self._source_traceback is not None:
-                    context['source_traceback'] = self._source_traceback
-                self._loop.call_exception_handler(context)
+    def __del__(self):
+        try:
+            _conn = self._conn
+        except AttributeError:
+            return
+        if _conn is not None and not _conn.closed:
+            self.close()
+            warnings.warn(f"Unclosed connection {self!r}",
+                          ResourceWarning)
+
+            context = {'connection': self,
+                       'message': 'Unclosed connection'}
+            if self._source_traceback is not None:
+                context['source_traceback'] = self._source_traceback
+            self._loop.call_exception_handler(context)
 
     @property
     def notifies(self):
         """Return notification queue."""
         return self._notifies
 
-    if PY_35:  # pragma: no branch
-        @asyncio.coroutine
-        def __aenter__(self):
-            return self
+    async def _get_oids(self):
+        cur = await self.cursor()
+        rv0, rv1 = [], []
+        try:
+            await cur.execute(
+                "SELECT t.oid, typarray "
+                "FROM pg_type t JOIN pg_namespace ns ON typnamespace = ns.oid "
+                "WHERE typname = 'hstore';"
+            )
 
-        @asyncio.coroutine
-        def __aexit__(self, exc_type, exc_val, exc_tb):
-            self.close()
+            async for oids in cur:
+                if isinstance(oids, Mapping):
+                    rv0.append(oids['oid'])
+                    rv1.append(oids['typarray'])
+                else:
+                    rv0.append(oids[0])
+                    rv1.append(oids[1])
+        finally:
+            cur.close()
+
+        return tuple(rv0), tuple(rv1)
+
+    async def _connect(self):
+        try:
+            await self._poll(self._waiter, self._timeout)
+        except BaseException:
+            await asyncio.shield(self.close())
+            raise
+        if self._enable_json:
+            psycopg2.extras.register_default_json(self._conn)
+        if self._enable_uuid:
+            psycopg2.extras.register_uuid(conn_or_curs=self._conn)
+        if self._enable_hstore:
+            oids = await self._get_oids()
+            if oids is not None:
+                oid, array_oid = oids
+                psycopg2.extras.register_hstore(
+                    self._conn,
+                    oid=oid,
+                    array_oid=array_oid
+                )
+
+        return self
+
+    def __await__(self):
+        return self._connect().__await__()
+
+    async def __aenter__(self):
+        return self
+
+    async def __aexit__(self, exc_type, exc_val, exc_tb):
+        await self.close()
